#!/usr/bin/perl -w

#
# JdbIO.pm
# $Id: IO.pm 522 2007-11-11 05:45:40Z johnh $
#
# Copyright (C) 2005-2007 by John Heidemann <johnh@isi.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#


package Jdb::IO;

=head1 NAME

Jdb::IO - base class for Jdb IO (JdbReader and JdbWriter)

=head1 FUNCTIONS

=cut

@ISA = ();
($VERSION) = ('$Revision: 522 $' =~ m/(\d+)/);

use strict;
use IO::File;
use Carp;

=head2 new

    $jdb = new Jdb::IO;

Creates a new IO object.  Usually you should not create a JdbIO object
directly, but instead create a C<JdbReader> or C<JdbWriter>.

=cut

sub new {
    my ($class, $file) = @_;
    my $self = bless {
	_fh => undef,	# filehandle to file

	_headerrow => undef,
	_header_set => undef,
	_fscode => undef,   # -F option code
        _fs => ' ',   # field separator
        _fsre => '\s+',   # field separator
	_empty => '-',

	_cols => [],    # array of names of the columns (fields)
	_cols_to_i => {},  # reverse hash mapping names to offsets

	_error => undef,  # error status
    }, $class;
    return $self;
}

=head2 config_one

    $jdb->config_one($arglist_aref);

Parse the first confiuration option on the list, removing it.
=cut

sub config_one {
    my($self, $aaref) = @_;
    if ($aaref->[0] eq '-fh') {
	shift @$aaref;
	$self->{_fh} = shift @$aaref;
    } elsif ($aaref->[0] eq '-header') {
	shift @$aaref;
	$self->{_headerrow} = shift @$aaref;
    } elsif ($aaref->[0] eq '-fscode') {
	shift @$aaref;
	my $code = shift @$aaref;
	$self->parse_fscode($code);
	$self->update_headerrow;
    } elsif ($aaref->[0] eq '-cols') {
	shift @$aaref;
	my $col_ref = $aaref->[0];  shift @$aaref;
	foreach (@$col_ref) {
	    $self->col_create($_);
	};
	$self->update_headerrow;
    } else {
	croak("unknown option: " . $aaref->[0]);
    };
}

=head2 config

    $jdb->config(-arg1 => $value1, -arg2 => $value2);

Parse all options in the list.
=cut

sub config {
    my($self) = shift @_;
    my(@args) = @_;
    while ($#args >= 0) {
	$self->config_one(\@args);
    };
}

=head2 close

    $jdb->close;

Closes the file, frees open file handle.
=cut

sub close {
    my($self) = @_;
    return if ($self->{_error});
    $self->{_fh}->close;
    $self->{_error} = 'closed';
}

=head2 error

    $jdb->error;

Returns a descrptive string if there is an error,
or undef if not.
=cut

sub error {
    my($self) = @_;
    return $self->{_error};
}

=head2 update_headerrow

internal: create the header the internal schema

=cut
sub update_headerrow {
    my $self = shift @_;
    my $h = "#h ";
    if ($self->{_fscode}) {
	$h .= "-F" . $self->{_fscode} . " ";
    };
    $h .= join(" ", @{$self->{_cols}});
    $self->{_headerrow} = $h;
}


=head2 parse_headerrow

internal: interpet the header

=cut
sub parse_headerrow {
    my($self, $headerrow) = @_;
    $self->{_headerrow} = $headerrow;
    my(@f) = split(/\s+/, $headerrow);
    my $tag = shift @f;
    die "jdb list format not supported yet.\n" if ($tag eq "#L");
    if ($tag ne "#h") {
        $self->{_error} = "header line is not jdb format\n";
	return;
   };
   #
   # handle options
   #
   while ($#f >= 0 && $f[0] =~ /^-(.)(.*)/) {
       my($key, $value) = ($1, $2);
       shift @f;
       if ($key eq 'F') {
	   $self->parse_fscode($value);
        }
    };

    # create them!
    foreach (@f) {
	$self->col_create($_);
    };
}

=head2 parse_fscode

internal

=cut
sub parse_fscode {
    my $self = shift @_;
    my $code = shift @_;
    my ($fsre, $outfs);
    if (!defined($code) || $code eq 'D') {  # default
        $fsre = '\s+';  # "[ \t\n]+";
        $outfs = "\t";
	$code = undef;
    } elsif ($code eq 'S') {   # double space
        $fsre = '\s\s+';
        $outfs = "  ";
    } elsif ($code eq 't') {   # single tab
        $fsre = "\t";
        $outfs = "\t";
    } else {   # anything else
        $code = eval "qq{$code}";  # handle backslash expansion
	$fsre = "[$code]+";
	$outfs = $code;
    };
    $self->{_fscode} = $code;
    $self->{_fsre} = $fsre;
    $self->{_fs} = $outfs;
}


=head2 establish_new_col_mapping

internal

=cut
sub establish_new_col_mapping {
    my($self, $colname) = @_;

    my $coli = $#{$self->{_cols}} + 1;
    $self->{_cols}->[$coli] = $colname;
    $self->{_cols_to_i}->{$colname} = $coli;
}

=head2 col_create

    $jdb->col_create($col_name)

Add a new column named $COL_NAME to the schema.
Returns undef on failure.

=cut

sub col_create {
    my($self, $colname) = @_;
    if ($self->{_header_set}) {
	$self->{_error} = "attempt to add column to frozen jdb handle (reader or writer that's been written to): $colname.\n";
	return undef;
    };
    if (defined($self->col_to_i($colname))) {
	$self->{_error} = "duplicate col definition: $colname.\n";
	return undef;
    };
    $self->establish_new_col_mapping($colname);
    return 1;
}


=head2 cols

    @fields = $jdb->cols;

Returns the column headings (the field names) of the open database.

=cut

sub cols {
    my($self) = @_;
    return $self->{_cols};
}


=head2 col_to_i

    @fields = $jdb->col_to_i($column_name);

Returns the column index (0-based) of a given $COLUMN_NAME.

=cut

sub col_to_i {
    my($self, $n) = @_;
    return $self->{_cols_to_i}->{$n};
}


1;
