#!/usr/bin/perl -w

#
# Jdb::IO::Reader.pm
# $Id: Reader.pm 522 2007-11-11 05:45:40Z johnh $
#
# Copyright (C) 2005 by John Heidemann <johnh@isi.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#


package Jdb::IO::Reader;

=head1 NAME

Jdb::IO::Reader - handle formatting reading from a jdb file

=cut

@ISA = qw(Jdb::IO);
($VERSION) = ('$Revision: 522 $' =~ m/(\d+)/);

use strict;
use IO::File;
use Carp;
use Jdb::IO;

=head2 new

    $jdb = new Jdb::IO::Reader(-file => $filename);
    $jdb = new Jdb::IO::Reader(-header => "#h -Ft foo bar", -fh => $file_handle);

Creates a new reader object from FILENAME.
(FILENAME can also be a IO::Handle object.)
Always succeeds, but 
check the C<error> method to test for failure.

Options:

=over

=item -fh FILE_HANDLE
Read IO from the given file handle.

=item -file FILENAME
Open and read the given filename.

=item -header HEADER_LINE
Force the header to the given HEADER_LINE

=back

User-specified -header arguments override a header provided in the input source.

=cut

sub new {
    my $class = shift @_;
    my $self = $class->SUPER::new(@_);
    bless $self, $class;
    #
    # new instance variables
    $self->{_unreadq} = [];
    #
    $self->config(@_);
    #
    # setup:    
    if (!$self->{_fh}) {
	$self->{_error} = "cannot setup filehandle";
	return $self;
    };

    $self->read_headerrow if (!defined($self->{_headerrow}));;
    $self->parse_headerrow($self->{_headerrow});
    $self->{_header_set} = 1;   # go read-only

    return $self;
}

=head2 config_one

documented in new

=cut
sub config_one {
    my($self, $aaref) = @_;
    if ($aaref->[0] eq '-file') {
	shift @$aaref;
	my($file) = shift @$aaref;
	my $fh = new IO::File $file, "r";
	if ($fh) {
	    $self->{_fh} = $fh;
	} else {
	    $self->{_error} = "cannot open $file";
	};
    } else {
	$self->SUPER::config_one($aaref);
    };
}


=head2 read_headerrow

internal use only; reads the header

=cut
sub read_headerrow {
    my($self) = @_;
    my $headerrow = $self->{_fh}->getline;
    if (!defined($headerrow)) {
	$self->{_error} = "no header line\n";
	return;
    };
    chomp $headerrow;
    $self->{_headerrow} = $headerrow;
};


=head2 read_row_to_aref

    $jdb->read_row_to_aref(\@a);

Then $a[0] is the 0th column, etc.
Returns undef if the read fails, typically due to EOF.

=cut

sub read_row_to_aref {
    my($self, $aref) = @_;

    return undef if (defined($self->{_error}));

    # first, check unread
    if ($#{$self->{_unreadq}} >= 0) {
	my $frontref = shift @{$self->{_unreadq}};
	@$aref = @$frontref;
	return 1;
    };

    my $line;
    LINE: while (1) {
	$line = $self->{_fh}->getline;
	if (!defined($line)) {
	    @$aref = ();
	    return undef;
	};
	chomp $line;
	last if ($line !~ /^\s*\#/);  # skip comments|
    };
    @$aref =  split(/$self->{_fsre}/, $line);
    return 1;
}

=head2 unread_row_from_aref

    $jdb->unread_row_from_aref(\%a);

Put array %a back into the file.

=cut

sub unread_row_from_aref {
    my($self, $aref) = @_;
    my @a = @$aref;  # make a copy
    unshift @{$self->{_unreadq}}, \@a;
}

=head2 read_row_to_href

    $jdb->read_row_to_href(\%h);

Read the next row into hash C<%h>.
Then $h{'colname'} is the value of that column.
Returns undef if the read fails, typically due to EOF.

=cut

sub read_row_to_href {
    my($self, $href) = @_;
    my @a;
    $self->read_row_to_aref(\@a) or return undef;
    foreach my $i (0..$#{$self->{_cols}}) {
	$href->{$self->{_cols}[$i]} = $a[$i];
    };
    return 1;
}

=head2 unread_row_from_href

    $jdb->unread_row_from_href(\%h);

Put hash %h back into the file.

=cut

sub unread_row_from_href {
    my($self, $href) = @_;
    my @a = ('-' x $#{$self->{_cols}});  # null record
    foreach (keys %$href) {
	my($i) = $self->{_cols_to_i}->{$_};
	defined($i) or croak "column name $_ is not in current file";
	$a[$i] = $href->{$_};
    };
    $self->unread_row_from_aref(\@a);
}


1;
